@@ -36,20 +36,18 @@ In addition, while many builtin functions are supported, several builtin
 functions that are considered unsafe are missing ('eval', 'exec', and
 'getattr' for example)
 """
-
-from __future__ import division, print_function
-
 import ast
 import time
 import inspect
 from sys import exc_info, stdout, stderr, version_info
 
-import six
-
 from .astutils import (UNSAFE_ATTRS, HAS_NUMPY, make_symbol_table, numpy,
                        op2func, ExceptionHolder, ReturnedNone,
                        valid_symbol_name)
 
+if version_info[0] < 3 or version_info[1] < 5:
+    raise SystemError("Python 3.5 or higher required")
+
 builtins = __builtins__
 if not isinstance(builtins, dict):
     builtins = builtins.__dict__
@@ -59,10 +57,8 @@ ALL_NODES = ['arg', 'assert', 'assign', 'attribute', 'augassign', 'binop',
              'dict', 'ellipsis', 'excepthandler', 'expr', 'extslice',
              'for', 'functiondef', 'if', 'ifexp', 'index', 'interrupt',
              'list', 'listcomp', 'module', 'name', 'nameconstant', 'num',
-             'pass', 'print', 'raise', 'repr', 'return', 'slice', 'str',
-             'subscript', 'try', 'tuple', 'unaryop', 'while']
-
-ERR_MAX_TIME = "Execution exceeded time limit, max runtime is {}s"
+             'pass', 'raise', 'repr', 'return', 'slice', 'str',
+             'subscript', 'try', 'tuple', 'unaryop', 'while', 'constant']
 
 class Interpreter(object):
     """create an asteval Interpreter: a restricted, simplified interpreter
@@ -106,14 +102,15 @@ class Interpreter(object):
         whether to support `raise`.
     no_print : bool
         whether to support `print`.
-    max_time : float
-        deprecated.  max run time in seconds (see Note 2) [30.0]
+    readonly_symbols : iterable or `None`
+        symbols that the user can not assign to
+    builtins_readonly : bool
+        whether to blacklist all symbols that are in the initial symtable
 
     Notes
     -----
     1. setting `minimal=True` is equivalent to setting all
        `no_***` options to `True`.
-    2. max_time is not reliable and support may be dropped soon.
     """
 
     def __init__(self, symtable=None, usersyms=None, writer=None,
@@ -121,7 +118,8 @@ class Interpreter(object):
                  no_if=False, no_for=False, no_while=False, no_try=False,
                  no_functiondef=False, no_ifexp=False, no_listcomp=False,
                  no_augassign=False, no_assert=False, no_delete=False,
-                 no_raise=False, no_print=False, max_time=30):
+                 no_raise=False, no_print=False, max_time=None,
+                 readonly_symbols=None, builtins_readonly=False):
 
         self.writer = writer or stdout
         self.err_writer = err_writer or stderr
@@ -131,8 +129,6 @@ class Interpreter(object):
                 usersyms = {}
             symtable = make_symbol_table(use_numpy=use_numpy, **usersyms)
 
-        symtable['print'] = self._printer
-
         self.symtable = symtable
         self._interrupt = None
         self.error = []
@@ -141,9 +137,11 @@ class Interpreter(object):
         self.retval = None
         self.lineno = 0
         self.start_time = time.time()
-        self.max_time = max_time
         self.use_numpy = HAS_NUMPY and use_numpy
 
+        symtable['print'] = self._printer
+        self.no_print = no_print or minimal
+
         nodes = ALL_NODES[:]
 
         if minimal or no_if:
@@ -164,8 +162,6 @@ class Interpreter(object):
             nodes.remove('delete')
         if minimal or no_raise:
             nodes.remove('raise')
-        if minimal or no_print:
-            nodes.remove('print')
         if minimal or no_listcomp:
             nodes.remove('listcomp')
         if minimal or no_augassign:
@@ -180,6 +176,14 @@ class Interpreter(object):
             self.node_handlers['tryexcept'] = self.node_handlers['try']
             self.node_handlers['tryfinally'] = self.node_handlers['try']
 
+        if readonly_symbols is None:
+            self.readonly_symbols = set()
+        else:
+            self.readonly_symbols = set(readonly_symbols)
+
+        if builtins_readonly:
+            self.readonly_symbols |= set(self.symtable)
+
         self.no_deepcopy = [key for key, val in symtable.items()
                             if (callable(val)
                                 or 'numpy.lib.index_tricks' in repr(val)
@@ -234,12 +238,12 @@ class Interpreter(object):
         if len(self.error) > 0 and not isinstance(node, ast.Module):
             msg = '%s' % msg
         err = ExceptionHolder(node, exc=exc, msg=msg, expr=expr, lineno=lineno)
-        self._interrupt = ast.Break()
+        self._interrupt = ast.Raise()
         self.error.append(err)
         if self.error_msg is None:
-            self.error_msg = "%s in expr='%s'" % (msg, self.expr)
+            self.error_msg = "at expr='%s'" % (self.expr)
         elif len(msg) > 0:
-            self.error_msg = "%s\n %s" % (self.error_msg, msg)
+            self.error_msg = msg
         if exc is None:
             try:
                 exc = self.error[0].exc
@@ -267,11 +271,13 @@ class Interpreter(object):
         """Execute parsed Ast representation for an expression."""
         # Note: keep the 'node is None' test: internal code here may run
         #    run(None) and expect a None in return.
-        if time.time() - self.start_time > self.max_time:
-            raise RuntimeError(ERR_MAX_TIME.format(self.max_time))
         out = None
         if len(self.error) > 0:
             return out
+        if self.retval is not None:
+            return self.retval
+        if isinstance(self._interrupt, (ast.Break, ast.Continue)):
+            return self._interrupt
         if node is None:
             return out
         if isinstance(node, str):
@@ -297,7 +303,10 @@ class Interpreter(object):
             return ret
         except:
             if with_raise:
-                self.raise_exception(node, expr=expr)
+                if len(self.error) == 0:
+                    # Unhandled exception that didn't go through raise_exception
+                    self.raise_exception(node, expr=expr)
+                raise
 
     def __call__(self, expr, **kw):
         """Call class instance as function."""
@@ -398,7 +407,8 @@ class Interpreter(object):
     def on_assert(self, node):    # ('test', 'msg')
         """Assert statement."""
         if not self.run(node.test):
-            self.raise_exception(node, exc=AssertionError, msg=node.msg)
+            msg = node.msg.s if node.msg else ""
+            self.raise_exception(node, exc=AssertionError, msg=msg)
         return True
 
     def on_list(self, node):    # ('elt', 'ctx')
@@ -414,6 +424,10 @@ class Interpreter(object):
         return dict([(self.run(k), self.run(v)) for k, v in
                      zip(node.keys, node.values)])
 
+    def on_constant(self, node):   # ('value', 'kind')
+        """Return constant value."""
+        return node.value
+
     def on_num(self, node):   # ('n',)
         """Return number."""
         return node.n
@@ -450,7 +464,7 @@ class Interpreter(object):
 
         """
         if node.__class__ == ast.Name:
-            if not valid_symbol_name(node.id):
+            if not valid_symbol_name(node.id) or node.id in self.readonly_symbols:
                 errmsg = "invalid symbol name (reserved word?) %s" % node.id
                 self.raise_exception(node, exc=NameError, msg=errmsg)
             self.symtable[node.id] = val
@@ -469,6 +483,8 @@ class Interpreter(object):
             xslice = self.run(node.slice)
             if isinstance(node.slice, ast.Index):
                 sym[xslice] = val
+            elif isinstance(node.slice, ast.Constant):
+                sym[xslice] = val
             elif isinstance(node.slice, ast.Slice):
                 sym[slice(xslice.start, xslice.stop)] = val
             elif isinstance(node.slice, ast.ExtSlice):
@@ -535,9 +551,9 @@ class Interpreter(object):
         nslice = self.run(node.slice)
         ctx = node.ctx.__class__
         if ctx in (ast.Load, ast.Store):
-            if isinstance(node.slice, (ast.Index, ast.Slice, ast.Ellipsis)):
+            if isinstance(node.slice, (ast.Index, ast.Constant, ast.Slice, ast.Ellipsis)):
                 return val.__getitem__(nslice)
-            elif isinstance(node.slice, ast.ExtSlice):
+            elif isinstance(node.slice, (ast.ExtSlice, ast.UnaryOp)):
                 return val[nslice]
         else:
             msg = "subscript with unknown context"
@@ -553,7 +569,7 @@ class Interpreter(object):
                 children.append(tnode.attr)
                 tnode = tnode.value
 
-            if tnode.__class__ == ast.Name:
+            if tnode.__class__ == ast.Name and tnode.id not in self.readonly_symbols:
                 children.append(tnode.id)
                 children.reverse()
                 self.symtable.pop('.'.join(children))
@@ -582,36 +598,28 @@ class Interpreter(object):
         return val
 
     def on_compare(self, node):  # ('left', 'ops', 'comparators')
-        """comparison operators"""
+        """comparison operators, including chained comparisons (a<b<c)"""
         lval = self.run(node.left)
-        out = True
+        results = []
         for op, rnode in zip(node.ops, node.comparators):
             rval = self.run(rnode)
-            out = op2func(op)(lval, rval)
+            ret = op2func(op)(lval, rval)
+            results.append(ret)
+            if (self.use_numpy and not isinstance(ret, numpy.ndarray)) and not ret:
+                break
             lval = rval
-            if self.use_numpy and isinstance(out, numpy.ndarray) and out.any():
-                break
-            elif not out:
-                break
+        if len(results) == 1:
+            return results[0]
+        else:
+            out = True
+            for r in results:
+                out = out and r
         return out
 
-    def on_print(self, node):    # ('dest', 'values', 'nl')
-        """Note: implements Python2 style print statement, not print()
-        function.
-
-        May need improvement....
-
-        """
-        dest = self.run(node.dest) or self.writer
-        end = ''
-        if node.nl:
-            end = '\n'
-        out = [self.run(tnode) for tnode in node.values]
-        if out and len(self.error) == 0:
-            self._printer(*out, file=dest, end=end)
-
     def _printer(self, *out, **kws):
         """Generic print function."""
+        if self.no_print:
+            return
         flush = kws.pop('flush', True)
         fileh = kws.pop('file', self.writer)
         sep = kws.pop('sep', ' ')
@@ -704,6 +712,7 @@ class Interpreter(object):
                         for tline in hnd.body:
                             self.run(tline)
                         break
+                break
         if no_errors and hasattr(node, 'orelse'):
             for tnode in node.orelse:
                 self.run(tnode)
@@ -714,12 +723,8 @@ class Interpreter(object):
 
     def on_raise(self, node):    # ('type', 'inst', 'tback')
         """Raise statement: note difference for python 2 and 3."""
-        if version_info[0] == 3:
-            excnode = node.exc
-            msgnode = node.cause
-        else:
-            excnode = node.type
-            msgnode = node.inst
+        excnode = node.exc
+        msgnode = node.cause
         out = self.run(excnode)
         msg = ' '.join(out.args)
         msg2 = self.run(msgnode)
@@ -741,13 +746,16 @@ class Interpreter(object):
             args = args + self.run(starargs)
 
         keywords = {}
-        if six.PY3 and func == print:
+        if func == print:
             keywords['file'] = self.writer
-
         for key in node.keywords:
             if not isinstance(key, ast.keyword):
                 msg = "keyword error in function call '%s'" % (func)
                 self.raise_exception(node, msg=msg)
+            if key.arg in keywords:
+                self.raise_exception(node,
+                                     msg="keyword argument repeated: %s" % key.arg,
+                                     exc=SyntaxError)                
             keywords[key.arg] = self.run(key.value)
 
         kwargs = getattr(node, 'kwargs', None)
@@ -756,8 +764,11 @@ class Interpreter(object):
 
         try:
             return func(*args, **keywords)
-        except:
-            self.raise_exception(node, msg="Error running %s" % (func))
+        except Exception as ex:
+            func_name = getattr(func, '__name__', str(func))
+            self.raise_exception(
+                node, msg="Error running function call '%s' with args %s and "
+                "kwargs %s: %s" % (func_name, args, keywords, ex))
 
     def on_arg(self, node):    # ('test', 'msg')
         """Arg for function definitions."""
@@ -770,17 +781,17 @@ class Interpreter(object):
             raise Warning("decorated procedures not supported!")
         kwargs = []
 
+        if not valid_symbol_name(node.name) or node.name in self.readonly_symbols:
+            errmsg = "invalid function name (reserved word?) %s" % node.name
+            self.raise_exception(node, exc=NameError, msg=errmsg)
+
         offset = len(node.args.args) - len(node.args.defaults)
         for idef, defnode in enumerate(node.args.defaults):
             defval = self.run(defnode)
             keyval = self.run(node.args.args[idef+offset])
             kwargs.append((keyval, defval))
 
-        if version_info[0] == 3:
-            args = [tnode.arg for tnode in node.args.args[:offset]]
-        else:
-            args = [tnode.id for tnode in node.args.args[:offset]]
-
+        args = [tnode.arg for tnode in node.args.args[:offset]]
         doc = None
         nb0 = node.body[0]
         if isinstance(nb0, ast.Expr) and isinstance(nb0.value, ast.Str):
@@ -788,11 +799,10 @@ class Interpreter(object):
 
         varkws = node.args.kwarg
         vararg = node.args.vararg
-        if version_info[0] == 3:
-            if isinstance(vararg, ast.arg):
-                vararg = vararg.arg
-            if isinstance(varkws, ast.arg):
-                varkws = varkws.arg
+        if isinstance(vararg, ast.arg):
+            vararg = vararg.arg
+        if isinstance(varkws, ast.arg):
+            varkws = varkws.arg
 
         self.symtable[node.name] = Procedure(node.name, self, doc=doc,
                                              lineno=self.lineno,
@@ -817,6 +827,7 @@ class Procedure(object):
         """TODO: docstring in public method."""
         self.__ininit__ = True
         self.name = name
+        self.__name__ = self.name
         self.__asteval__ = interp
         self.raise_exc = self.__asteval__.raise_exception
         self.__doc__ = doc
